import re


class ScalaToPyspark:

    def __init__(self, scala_path):
        self.scala_path = scala_path

    def lines(self):
        text_file = open(self.scala_path, "r")
        result = text_file.readlines()

        # remove lines that start with package
        result = list(filter(lambda l: not l.startswith("package"), result))

        # replace common import statements
        result = ["from pyspark.sql.dataframe import DataFrame\n" if x.startswith("import org.apache.spark.sql.DataFrame") else x for x in result]
        result = ["import pyspark.sql.functions as F\n" if x.startswith("import org.apache.spark.sql.functions") else x for x in result]

        # remove scala import statements
        result = list(filter(lambda l: not l.startswith("import scala"), result))

        # remove the vals and vars
        result = [x.replace("val ", "") for x in result]
        result = [x.replace("var ", "") for x in result]

        # remove ending curly braces (low hanging fruit)
        result = list(filter(lambda l: l.strip() != "}", result))

        # clean up the function definitions
        result = [self.clean_function_definition(x) if x.lstrip().startswith("def") else x for x in result]

        # prefix SQL functions with F
        result = [x.replace("col(", "F.col(") for x in result]

        # replace null with None
        result = [x.replace("null", "None") for x in result]

        # capital case true and false
        result = [x.replace("true", "True") for x in result]
        result = [x.replace("false", "False") for x in result]

        return result

    def display(self):
        print("".join(self.lines()))

    def clean_function_definition(self, s):
        m = re.search('(\s*)def (\w+).*\((.*)\)', s)
        if m:
            whitespace = m.group(1)
            method_name = m.group(2)
            args = m.group(3)
            cleaned_args = self.clean_args(args)
            return "{whitespace}def {method_name}({cleaned_args}):\n".format(
                whitespace=whitespace,
                method_name=method_name,
                cleaned_args=cleaned_args
            )
        return s

    def clean_args(self, args):
        unannoted_args_list = list(map(lambda a: a.split(": ")[0], args.split(", ")))
        return ", ".join(unannoted_args_list)



def column_to_list(df, col_name):
    return [x[col_name] for x in df.select(col_name).collect()]


def two_columns_to_dictionary(df, key_col_name, value_col_name):
    k, v = key_col_name, value_col_name
    return {x[k]: x[v] for x in df.select(k, v).collect()}


def to_list_of_dictionaries(df):
    return list(map(lambda r: r.asDict(), df.collect()))


def print_athena_create_table(df, athena_table_name, s3location):
    fields = df.schema

    print(f"CREATE EXTERNAL TABLE IF NOT EXISTS `{athena_table_name}` ( ")

    for field in fields.fieldNames()[:-1]:
        print("\t", f"`{fields[field].name}` {fields[field].dataType.simpleString()}, ")
    last = fields[fields.fieldNames()[-1]]
    print("\t", f"`{last.name}` {last.dataType.simpleString()} ")

    print(")")
    print("STORED AS PARQUET")
    print(f"LOCATION '{s3location}'\n")


def show_output_to_df(show_output, spark):
    l = show_output.split("\n")
    ugly_column_names = l[1]
    pretty_column_names = [i.strip() for i in ugly_column_names[1:-1].split("|")]
    pretty_data = []
    ugly_data = l[3:-1]
    for row in ugly_data:
        r = [i.strip() for i in row[1:-1].split("|")]
        pretty_data.append(tuple(r))
    return spark.createDataFrame(pretty_data, pretty_column_names)



from pyspark.sql.types import StructType, StructField
from pyspark.sql import SparkSession


def create_df(self, rows_data, col_specs):
    struct_fields = list(map(lambda x: StructField(*x), col_specs))
    return self.createDataFrame(data=rows_data, schema=StructType(struct_fields))


SparkSession.create_df = create_df



import pyspark.sql.functions as F


def with_columns_renamed(fun):
    def _(df):
        cols = list(map(
            lambda col_name: F.col("`{0}`".format(col_name)).alias(fun(col_name)),
            df.columns
        ))
        return df.select(*cols)
    return _


def with_some_columns_renamed(fun, change_col_name):
    def _(df):
        cols = list(map(
            lambda col_name: F.col("`{0}`".format(col_name)).alias(fun(col_name)) if change_col_name(col_name) else F.col("`{0}`".format(col_name)),
            df.columns
        ))
        return df.select(*cols)
    return _


def snake_case_col_names(df):
    return with_columns_renamed(to_snake_case)(df)


def to_snake_case(s):
    return s.lower().replace(" ", "_")


def sort_columns(df, sort_order):
    sorted_col_names = None
    if sort_order == "asc":
        sorted_col_names = sorted(df.columns)
    elif sort_order == "desc":
        sorted_col_names = sorted(df.columns, reverse=True)
    else:
        raise ValueError("['asc', 'desc'] are the only valid sort orders and you entered a sort order of '{sort_order}'".format(
            sort_order=sort_order
        ))
    return df.select(*sorted_col_names)




import pyspark.sql.functions as F

from pyspark.sql.types import BooleanType


def single_space(col):
    return F.trim(F.regexp_replace(col, " +", " "))


def remove_all_whitespace(col):
    return F.regexp_replace(col, "\\s+", "")


def anti_trim(col):
    return F.regexp_replace(col, "\\b\\s+\\b", "")


def remove_non_word_characters(col):
    return F.regexp_replace(col, "[^\\w\\s]+", "")


def exists(f):
    def temp_udf(l):
        return any(map(f, l))
    return F.udf(temp_udf, BooleanType())


def forall(f):
    def temp_udf(l):
        return all(map(f, l))
    return F.udf(temp_udf, BooleanType())


def multi_equals(value):
    def temp_udf(*cols):
        return all(map(lambda col: col == value, cols))
    return F.udf(temp_udf, BooleanType())


def week_start_date(col, week_start_day = 'Sun'):
    _raise_if_invalid_day(week_start_day)
    # the "standard week" in Spark is from Sunday to Saturday
    mapping = {'Sun': 'Sat', 'Mon': 'Sun', 'Tue': 'Mon', 'Wed': 'Tue', 'Thu': 'Wed', 'Fri': 'Thu', 'Sat': 'Fri'}
    end = week_end_date(col, mapping[week_start_day])
    return F.date_add(end, -6)


def week_end_date(col, week_end_day = 'Sat'):
    _raise_if_invalid_day(week_end_day)
    # these are the default Spark mappings.  Spark considers Sunday the first day of the week.
    day_of_week_mapping = {'Sun': 1, 'Mon': 2, 'Tue': 3, 'Wed': 4, 'Thu': 5, 'Fri': 6, 'Sat': 7}
    return F.when(F.dayofweek(col).eqNullSafe(F.lit(day_of_week_mapping[week_end_day])), col).otherwise(F.next_day(col, week_end_day))


def _raise_if_invalid_day(day):
    valid_days = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    if day not in valid_days:
        message = "The day you entered '{0}' is not valid.  Here are the valid days: [{1}]".format(day, ','.join(valid_days))
        raise ValueError(message)


def approx_equal(col1, col2, threshhold):
    return F.abs(col1 - col2) < threshhold

